// =============================================================
// File:	SNIa.js
// Purpose:	Type-Ia supernovae calculations
// Copyright:	(c) 2018 Henk Reints, MSc. http://henk-reints.nl
// Date:	2018-06-19
// =============================================================
var speedOfLightInMetersPerSecond	= 299792458;
var secondsPerYear			= 365.25*24*60*60;
var arcsecInRadians			= Math.PI/180/60/60;
var astronomicalUnitInMeters		= 149597870700;
var metersPerKilometer			= 1000;
var kilo				= 1e3;
var mega				= 1e6;
var giga				= 1e9;
var distanceOfAbsMagnitudeInParsecs	= 10;
var magnitudeInterval			= 5;
var fluxRatioPerMagnitudeInterval	= 100;
var distanceModulusToExponentFactor	= 0.5/magnitudeInterval;	// 0.5 for square root (surface area to distance conversion)
//
var maxDeltaHubbleConstantInKmPerSecPerMpc	= 0.001;	// stop criterium for iteration in 3-spherical Hubble Distance calculation
var maxIterationCount				= 100;		// to prevent race condition
//
var absoluteMagnitudeOfSNIa		= -19.3;		// as found on the Internet
//
var speedOfLightInKmPerSec		= speedOfLightInMetersPerSecond/metersPerKilometer;
var lightyearInMeters			= speedOfLightInMetersPerSecond*secondsPerYear;
var parsecInMeters			= astronomicalUnitInMeters/Math.atan(arcsecInRadians);
var MegaParsecInKilometers		= mega*parsecInMeters/metersPerKilometer;
//
var wikipediaHubbleConstantInKmPerSecPerMpc	= 73.52;	// latest value on https://en.wikipedia.org/wiki/Hubble%27s_law as of 2018-06-16
var basicHubbleConstantInKmPerSecPerMpc		= wikipediaHubbleConstantInKmPerSecPerMpc;
var HubbleConstantLowerDateBoundary		= "";		// obtain only Hubble Constant values published since this date
var HubbleConstantInKmPerSecPerMpc		= NaN;
var HubbleTimeInSeconds				= NaN;
var HubbleTimeInYears				= NaN;
var HubbleDistanceInMeters			= NaN;
var HubbleDistanceInLightyears			= NaN;
var HubbleDistanceInParsecs			= NaN;
//
function adjustHubbleParameters(argHubbleConstantInKmPerSecPerMpc)
{	HubbleConstantInKmPerSecPerMpc	= argHubbleConstantInKmPerSecPerMpc;
	HubbleTimeInSeconds		= MegaParsecInKilometers/HubbleConstantInKmPerSecPerMpc;
	HubbleTimeInYears		= HubbleTimeInSeconds/secondsPerYear;
	HubbleDistanceInMeters		= speedOfLightInMetersPerSecond*HubbleTimeInSeconds;
	HubbleDistanceInLightyears	= HubbleDistanceInMeters/lightyearInMeters;
	HubbleDistanceInParsecs		= HubbleDistanceInMeters/parsecInMeters;
};
function initHubbleParameters()
{	adjustHubbleParameters(basicHubbleConstantInKmPerSecPerMpc);
};
initHubbleParameters();
//
// redshiftList has been defined ad filled in the .wsf script
function getRedshift(supernovaId)
{	if (!redshiftList.indexSN.hasOwnProperty(supernovaId)) return NaN;
	var redshift = redshiftList[redshiftList.indexSN[supernovaId]].redshift;
	return toNumber(redshift);
};
function betaFromRedshift(argRedshift)
{	var zeta  = argRedshift+1;
	var zeta2 = zeta*zeta;
	var beta  = (zeta2-1)/(zeta2+1);
	return beta;
};
function distanceModulusToParsecsEuclidean(argDistanceModulus)
{	return distanceOfAbsMagnitudeInParsecs*Math.pow(fluxRatioPerMagnitudeInterval,argDistanceModulus*distanceModulusToExponentFactor);
};
function distanceModulusToParsecs3spherical(argDistanceModulus)
{	return Math.asin(distanceModulusToParsecsEuclidean(argDistanceModulus)*Math.PI/HubbleDistanceInParsecs)*HubbleDistanceInParsecs/Math.PI;
};
function computeEuclidean(argApparentMagnitude,argRedshift)
{	var beta		= betaFromRedshift(argRedshift);
	var velocityInKmPerSec	= beta*speedOfLightInKmPerSec;
	var distanceModulus	= argApparentMagnitude - absoluteMagnitudeOfSNIa;
	var distanceInParsecs	= distanceModulusToParsecsEuclidean(distanceModulus);
	var result =
		{geometry				: "Euclidean"
		,velocityInKmPerSec			: velocityInKmPerSec
		,distanceInParsecs			: distanceInParsecs
		,HubbleConstantInKmPerSecPerMpc		: velocityInKmPerSec*mega/distanceInParsecs
		,deltaHubbleConstantInKmPerSecPerMpc	: NaN
		,iterationCount				: NaN
		,iterationDidConverge			: null
		};
	return result;
};
function compute3spherical(argApparentMagnitude,argRedshift)
{	var beta		= betaFromRedshift(argRedshift);
	var velocityInKmPerSec	= beta*speedOfLightInKmPerSec;
	var distanceModulus	= argApparentMagnitude - absoluteMagnitudeOfSNIa;
	for (var iterationCount = 0, iterationDidConverge = false; iterationCount < maxIterationCount; iterationCount++)
	{	var oldHubbleConstantInKmPerSecPerMpc	= HubbleConstantInKmPerSecPerMpc;
		var distanceInParsecs			= distanceModulusToParsecs3spherical(distanceModulus);
		if (beta > 0.5) distanceInParsecs = HubbleDistanceInParsecs - distanceInParsecs;	// beta > 0.5 means it is in the far half of the Hubble distance
		adjustHubbleParameters(velocityInKmPerSec*mega/distanceInParsecs);
		var deltaHubbleConstantInKmPerSecPerMpc	= HubbleConstantInKmPerSecPerMpc - oldHubbleConstantInKmPerSecPerMpc;
		if (!isFinite(distanceInParsecs)) break;
		if (Math.abs(deltaHubbleConstantInKmPerSecPerMpc) < maxDeltaHubbleConstantInKmPerSecPerMpc)
		{	iterationDidConverge = true;
			break;
		};
	};
	var result =
		{geometry				: "3-spherical"
		,velocityInKmPerSec			: velocityInKmPerSec
		,distanceInParsecs			: distanceInParsecs
		,HubbleConstantInKmPerSecPerMpc		: HubbleConstantInKmPerSecPerMpc
		,deltaHubbleConstantInKmPerSecPerMpc	: deltaHubbleConstantInKmPerSecPerMpc
		,iterationCount				: iterationCount
		,iterationDidConverge			: iterationDidConverge
		};
	initHubbleParameters();	// restore what we have just ruined...
	return result;
};
// =================================================================================================
function filterHubbleConstant(record)
{	return record.publicationDate >= HubbleConstantLowerDateBoundary;
};
function filterRedshift(record)
{	var redshiftValue = toNumber(record.redshift);
	var result = isFinite(redshiftValue);
	return result;
};
function filterSupernova(record)
{	return isFinite(toNumber(record.Magn)) && isFinite(getRedshift(record.SN));
};
// =================================================================================================
var countDistanceEuclideanInParsecsAboveH0	= 0;
var countWithinSigmaH0Euclidean			= 0;
var countWithinSigmaH03spherical		= 0;
function processSupernova(supernovaId,magnitude,redshift)
{	var beta				= betaFromRedshift(redshift);
	var lightTravelFactor			= beta/(1+beta);
	var properDistanceInParsecs		= beta*HubbleDistanceInParsecs;
	var dataEuclidean			= computeEuclidean(magnitude,redshift);
	var data3spherical			= compute3spherical(magnitude,redshift);
	var result =
		{SN				: supernovaId
		,magnitude			: magnitude
		,redshift			: redshift
		,beta				: beta
		,lightTravelFactor		: lightTravelFactor
		,lookBackTimeInYears		: lightTravelFactor*HubbleTimeInYears
		,lightTravelDistanceInParsecs	: lightTravelFactor*HubbleDistanceInParsecs
		,properDistanceInParsecs	: properDistanceInParsecs
		,distanceEuclideanInParsecs	: dataEuclidean.distanceInParsecs
		,ratioEuclideanToProper		: dataEuclidean.distanceInParsecs/properDistanceInParsecs
		,distance3spherical0InParsecs	: distanceModulusToParsecs3spherical(magnitude-absoluteMagnitudeOfSNIa)
		,distance3spherical1InParsecs	: data3spherical.distanceInParsecs
		,H0Euclidean			: dataEuclidean.HubbleConstantInKmPerSecPerMpc
		,H03spherical			: data3spherical.HubbleConstantInKmPerSecPerMpc
		,iterationDidConverge		: data3spherical.iterationDidConverge
		};
	if (arguments.length == 0)
	{	var fieldNameList = [];
		for (var fieldName in result) if (result.hasOwnProperty(fieldName)) fieldNameList.push(fieldName);
		result = fieldNameList;
	} else
	{	if (dataEuclidean.distanceInParsecs > HubbleDistanceInParsecs) countDistanceEuclideanInParsecsAboveH0++;
	};
	return result;
};
var processedSupernovaFieldInfoList =
	{SN				: {caption:"SN"		,fieldWidth:6	,description:"Supernova ID"}
	,magnitude			: {caption:"magn."	,fieldWidth:4	,description:"apparent magnitude"}
	,redshift			: {caption:"redshift"	,fieldWidth:8	,description:"redshift"}
	,beta				: {caption:"beta"	,fieldWidth:8	,description:"redshift based velocity/speedOfLight (* = beta > 0.5)"}
	,lightTravelFactor		: {caption:"lTrvFctr"	,fieldWidth:8	,description:"redshift based light travel factor"}
	,lookBackTimeInYears		: {caption:"lookBack"	,fieldWidth:8	,description:"redshift based lookback time /year"}
	,lightTravelDistanceInParsecs	: {caption:"lTrvDist"	,fieldWidth:8	,description:"redshift based light travel distance /parsec"}
	,properDistanceInParsecs	: {caption:"propDist"	,fieldWidth:8	,description:"redshift based proper distance /parsec"}
	,distanceEuclideanInParsecs	: {caption:"distEucl"	,fieldWidth:8	,description:"Euclidean magnitude based distance /parsec (* = distEucl > Hubble distance)"}
	,ratioEuclideanToProper		: {caption:"Euc/Prop"	,fieldWidth:8	,description:"Euclidean magnitude based distance to redshift based proper distance ratio"}
	,distance3spherical0InParsecs	: {caption:"dist3sp0"	,fieldWidth:8	,description:"Initial  3-spherical magnitude based distance /parsec"}
	,distance3spherical1InParsecs	: {caption:"dist3sp1"	,fieldWidth:8	,description:"Iterated 3-spherical magnitude based distance /parsec"}
	,H0Euclidean			: {caption:"H0-Eucl"	,fieldWidth:7	,description:"calculated Euclidean   Hubble constant (* = within sigma)"}
	,H03spherical			: {caption:"H0-3sph"	,fieldWidth:7	,description:"calculated 3-spherical Hubble constant (* = within sigma)"}
	};
var processedSupernovaFieldSeparatorWidth = 2;
var processedSupernovaListPageSize = 50;
//
function buildLegend()
{	var result = ["Legend:"];
	var maxCaptionWidth = 0;
	for (var propertyName in processedSupernovaFieldInfoList)
	{	if (!processedSupernovaFieldInfoList.hasOwnProperty(propertyName)) continue;
		var processedSupernovaFieldInfo = processedSupernovaFieldInfoList[propertyName];
		var caption = processedSupernovaFieldInfo.caption;
		maxCaptionWidth = Math.max(caption.length,maxCaptionWidth);
	};
	for (var propertyName in processedSupernovaFieldInfoList)
	{	if (!processedSupernovaFieldInfoList.hasOwnProperty(propertyName)) continue;
		var processedSupernovaFieldInfo = processedSupernovaFieldInfoList[propertyName];
		result.push("    "+rpad(processedSupernovaFieldInfo.caption,maxCaptionWidth)+" : "+processedSupernovaFieldInfo.description);
	};
	return result.join("\n");
};
function buildCaptionLines()
{	var captionLine0 = [], captionLine1 = [];
	for (var propertyName in processedSupernovaFieldInfoList)
	{	if (!processedSupernovaFieldInfoList.hasOwnProperty(propertyName)) continue;
		var processedSupernovaFieldInfo = processedSupernovaFieldInfoList[propertyName];
		var fieldWidth	= processedSupernovaFieldInfo.fieldWidth;
		var caption	= processedSupernovaFieldInfo.caption;
		var underline	= rpad("",Math.max(caption.length,fieldWidth),"-");
		captionLine0.push(rpad(caption	,fieldWidth+processedSupernovaFieldSeparatorWidth));
		captionLine1.push(rpad(underline,fieldWidth+processedSupernovaFieldSeparatorWidth));
	};
	return [captionLine0.join("").replace(/\s+$/,""),captionLine1.join("").replace(/\s+$/,"")].join("\n");
};
function formatProcessedSupernovaRecord(processedSupernovaRecord)
{	var pr = processedSupernovaRecord, fi = processedSupernovaFieldInfoList, sw = processedSupernovaFieldSeparatorWidth;
	var basicH0lo = basicHubbleConstantInKmPerSecPerMpc - basicHubbleConstantInKmPerSecPerMpcStdDev;
	var basicH0hi = basicHubbleConstantInKmPerSecPerMpc + basicHubbleConstantInKmPerSecPerMpcStdDev;
	var withinSigmaH0Euclidean	= pr.H0Euclidean  > basicH0lo && pr.H0Euclidean  < basicH0hi;
	var withinSigmaH03spherical	= pr.H03spherical > basicH0lo && pr.H03spherical < basicH0hi;
	if (withinSigmaH0Euclidean ) countWithinSigmaH0Euclidean++;
	if (withinSigmaH03spherical) countWithinSigmaH03spherical++;
	var result =
		[rpad(pr.SN,fi.SN.fieldWidth+sw)
		,formatNumber(pr.magnitude			,-1	,""	,2	,0	,0	,fi.magnitude			.fieldWidth+sw)
		,formatNumber(pr.redshift			,-5	,""	,2	,0	,0	,fi.redshift			.fieldWidth+sw)
		,rpad(formatNumber(pr.beta			,-5	,""	,1	,0	,0)
			+(pr.beta > 0.5 ? "*" : "")							,fi.beta			.fieldWidth+sw)
		,formatNumber(pr.lightTravelFactor		,-5	,""	,1	,0	,0	,fi.lightTravelFactor		.fieldWidth+sw+1)
		,formatNumber(pr.lookBackTimeInYears		, 4	,""	,0	,6	,0	,fi.lookBackTimeInYears		.fieldWidth+sw)
		,formatNumber(pr.lightTravelDistanceInParsecs	, 4	,""	,0	,6	,0	,fi.lightTravelDistanceInParsecs.fieldWidth+sw)
		,formatNumber(pr.properDistanceInParsecs	, 4	,""	,0	,6	,0	,fi.properDistanceInParsecs	.fieldWidth+sw)
		,rpad(formatNumber(pr.distanceEuclideanInParsecs, 4	,""	,0	,6	,0)
			+(pr.distanceEuclideanInParsecs > HubbleDistanceInParsecs ? "*" : "")		,fi.distanceEuclideanInParsecs	.fieldWidth+sw)
		,formatNumber(pr.ratioEuclideanToProper		,-3	,""	,2	,0	,0	,fi.ratioEuclideanToProper	.fieldWidth+sw)
		,formatNumber(pr.distance3spherical0InParsecs	, 4	,""	,0	,6	,0	,fi.distance3spherical0InParsecs.fieldWidth+sw)
		,formatNumber(pr.distance3spherical1InParsecs	, 4	,""	,0	,6	,0	,fi.distance3spherical1InParsecs.fieldWidth+sw-1)
		,rpad(formatNumber(pr.H0Euclidean		,-2	,""	,3	,6	,0)
			+(withinSigmaH0Euclidean ? "*" : "")						,fi.H0Euclidean			.fieldWidth+sw)
		,rpad(formatNumber(pr.H03spherical		,-2	,""	,3	,6	,0)
			+(withinSigmaH03spherical ? "*" : "")						,fi.H03spherical		.fieldWidth+sw)
		].join("").replace(/\s+$/,"");
	return result;
};
function formatSummaryLines(count)
{	var fi = processedSupernovaFieldInfoList, sw = processedSupernovaFieldSeparatorWidth;
	var summaryLine1 =
		[rpad("Maxima:"										,fi.SN				.fieldWidth+sw
													+fi.magnitude			.fieldWidth+sw)
		,formatNumber(greatestRedshift			,-5	,""	,2	,0	,0	,fi.redshift			.fieldWidth+sw)
		,formatNumber(greatestBeta			,-5	,""	,1	,0	,0	,fi.beta			.fieldWidth+sw)
		,rpad(""										,fi.lightTravelFactor		.fieldWidth+sw+1
													+fi.lookBackTimeInYears		.fieldWidth+sw
													+fi.lightTravelDistanceInParsecs.fieldWidth+sw
													+fi.properDistanceInParsecs	.fieldWidth+sw)
		,formatNumber(greatestDistanceEuclideanInParsecs, 4	,""	,0	,0	,0	,fi.distanceEuclideanInParsecs	.fieldWidth+sw)
		].join("").replace(/\s+$/,"");
	var summaryLine2 =
		[rpad("#within sigma:"									,fi.SN				.fieldWidth+sw
													+fi.magnitude			.fieldWidth+sw
													+fi.redshift			.fieldWidth+sw
													+fi.beta			.fieldWidth+sw
													+fi.lightTravelFactor		.fieldWidth+sw
													+fi.lookBackTimeInYears		.fieldWidth+sw
													+fi.lightTravelDistanceInParsecs.fieldWidth+sw
													+fi.properDistanceInParsecs	.fieldWidth+sw
													+fi.distanceEuclideanInParsecs	.fieldWidth+sw
													+fi.ratioEuclideanToProper	.fieldWidth+sw
													+fi.distance3spherical0InParsecs.fieldWidth+sw
													+fi.distance3spherical1InParsecs.fieldWidth+sw)
		,rpad(lpad(countWithinSigmaH0Euclidean	,6)						,fi.H0Euclidean			.fieldWidth+sw)
		,rpad(lpad(countWithinSigmaH03spherical	,6)						,fi.H03spherical		.fieldWidth+sw)
		].join("").replace(/\s+$/,"");
	result =
		[buildCaptionLines().split("\n")[1]
		,summaryLine1
		,summaryLine2
		,"#SN: "+count
		].join("\n");

	countWithinSigmaH0Euclidean
	countWithinSigmaH03spherical

	return result;
};
// supernovaList has been defined ad filled in the .wsf script
function processSupernovaList()
{	var result = [];
	result.fieldNameList = processSupernova();
	result.recordToString = function()
	{	var result = [];
		for (var propertyName in this) if (this.hasOwnProperty(propertyName)) result.push([propertyName,this[propertyName]].join(":"));
		return "{"+result+"}";
	};
	result.toString = function()
	{	var result = [];
		for (var index = 0; index < this.length; index++) result.push(this[index].toString());
		return "["+result.join("\n,")+"\n]";
	};
	countDistanceEuclideanInParsecsAboveH0	= 0;	// no 'var', these are global variables
	countWithinSigmaH0Euclidean		= 0;
	countWithinSigmaH03spherical		= 0;
	for (var supernovaListIndex = 0, supernovaCount = supernovaList.length; supernovaListIndex < supernovaCount; supernovaListIndex++)
	{	var supernovaRecord = supernovaList[supernovaListIndex];
		var supernovaId	= supernovaRecord.SN;
		var magnitude	= toNumber(supernovaRecord.Magn);
		var redshift	= getRedshift(supernovaId);
		var resultRecord = processSupernova(supernovaId,magnitude,redshift);
		resultRecord.toString = result.recordToString;
		result.push(resultRecord);
	};
	return result;
};
function formatProcessedSupernovaList(processedSupernovaList)
{	var captionLines = buildCaptionLines();
	var result = [];
	for (var processedSupernovaListIndex = 0, L = processedSupernovaList.length; processedSupernovaListIndex < L; processedSupernovaListIndex++)
	{	if (processedSupernovaListIndex%processedSupernovaListPageSize == 0)
		{	if (processedSupernovaListIndex > 0) result.push("");
			result.push(captionLines);
		};
		result.push(formatProcessedSupernovaRecord(processedSupernovaList[processedSupernovaListIndex]));
	};
	return result.join("\n");
};
function formatStatistics(name,stats,precision)
{	var result =
		["Statistics of "+name+": "
		,"mu: "		+	formatNumber(stats.mu		,precision,"",3,7)
		,", sigma: "	+	formatNumber(stats.sigma	,precision,"",3,7)
		,", median: "	+	formatNumber(stats.median	,precision,"",3,7)
		,", min: "	+	formatNumber(stats.min		,precision,"",3,7)
		,", max: "	+	formatNumber(stats.max		,precision,"",3,7)
		,", count: "	+	formatNumber(stats.count	,0)
		];
	return result.join("");
};
// ==================================================================================
// End of file - SNIa.js - Copyright (c) 2018 Henk Reints, MSc. http://henk-reints.nl
