// Copyright (c) 2017 Henk Reints, http://henk-reints.nl
// #####################################################################################################################
// Library of Mathematical stuff.
// =====================================================================================================================
HRmath = {};	// the bare object
// =====================================================================================================================
// Replication of the Math object so we don't need to directly use the Math object anywhere.
// See: https://developer.mozilla.org/en/docs/Web/JavaScript/Reference/Global_Objects/Math
// If required properties or Methods can easily be replaced.
// =====================================================================================================================
// Properties:
// -----------
HRmath.E	= Math.E;	// Euler's constant and the base of natural logarithms, approximately 2.718.
HRmath.LN2	= Math.LN2;	// Natural logarithm of 2, approximately 0.693.
HRmath.LN10	= Math.LN10;	// Natural logarithm of 10, approximately 2.303.
HRmath.LOG2E	= Math.LOG2E;	// Base 2 logarithm of E, approximately 1.443.
HRmath.LOG10E	= Math.LOG10E;	// Base 10 logarithm of E, approximately 0.434.
HRmath.PI	= Math.PI;	// Ratio of the circumference of a circle to its diameter, approximately 3.14159.
HRmath.SQRT1_2	= Math.SQRT1_2;	// Square root of 1/2; equivalently, 1 over the square root of 2, approximately 0.707.
HRmath.SQRT2	= Math.SQRT2;	// Square root of 2, approximately 1.414.
// =====================================================================================================================
// Methods:
// --------
HRmath.abs	= Math.abs;	// absolute value
HRmath.acos	= Math.acos;	// arccosine
HRmath.asin	= Math.asin;	// arcsine
HRmath.atan	= Math.atan;	// arctangent
HRmath.atan2	= Math.atan2;	// arctangent of the quotient of the arguments
HRmath.ceil	= Math.ceil;	// smallest integer greater than or equal
HRmath.cos	= Math.cos;	// cosine
HRmath.exp	= Math.exp;	// E^x
HRmath.floor	= Math.floor;	// largest integer less than or equal
HRmath.log	= Math.log;	// natural logarithm
HRmath.max	= Math.max;	// largest of zero or more numbers
HRmath.min	= Math.min;	// smallest of zero or more numbers
HRmath.pow	= Math.pow;	// base to the exponent power
HRmath.random	= Math.random;	//  pseudo-random number between 0 (inclusive) and 1 (exclusive)
HRmath.round	= Math.round;	// nearest integer
HRmath.sin	= Math.sin;	// sine
HRmath.sqrt	= Math.sqrt;	// positive square root
HRmath.tan	= Math.tan;	// tangent
// =====================================================================================================================
// (New) ECMAScript 2015 (6th edition) Methods
// implemented in a backward compatible way for older scripting engines like Windows Script Host
// ---------------------------------------------------------------------------------------------
// (Inverse) Hyperbolic functions:
// -------------------------------
HRmath.sinh	= typeof Math.sinh	== "function"	? Math.sinh	: function(x){return(HRmath.exp(x)-HRmath.exp(-x))/2};
HRmath.cosh	= typeof Math.cosh	== "function"	? Math.cosh	: function(x){return(HRmath.exp(x)+HRmath.exp(-x))/2};
HRmath.tanh	= typeof Math.tanh	== "function"	? Math.tanh	: function(x){var y=HRmath.exp(2*x);return(y-1)/(y+1)};
HRmath.asinh	= typeof Math.asinh	== "function"	? Math.asinh	: function(x){return HRmath.log(x+HRmath.sqrt(x*x+1))};
HRmath.acosh	= typeof Math.acosh	== "function"	? Math.acosh	: function(x){return HRmath.log(x+HRmath.sqrt(x*x-1))};
HRmath.atanh	= typeof Math.atanh	== "function"	? Math.atanh	: function(x){return 0.5*HRmath.log((1+x)/(1-x))};
// ----------------------
// Logarithm/exponential:
// ----------------------
HRmath.log10	= typeof Math.log10	== "function"	? Math.log10	: function(x){return HRmath.log(x)/HRmath.LN10};// base 10 logarithm of x
HRmath.log2	= typeof Math.log2	== "function"	? Math.log2	: function(x){return HRmath.log(x)/HRmath.LN2};	// base  2 logarithm of x
HRmath.log1p	= typeof Math.log1p	== "function"	? Math.log1p	: function(x){return HRmath.log(1+x)};		// natural logarithm of 1+x
HRmath.expm1	= typeof Math.expm1	== "function"	? Math.expm1	: function(x){return HRmath.exp(x)-1};		// exp(x)-1
// ------
// Other:
// ------
HRmath.sign	= typeof Math.sign	== "function"	? Math.sign	: function(x){return x<0?-1:x>0?1:0};		// sign of x
HRmath.cbrt	= typeof Math.cbrt	== "function"	? Math.cbrt	: function(x){return HRmath.root(x,3)};		// cube root of x
//
HRmath.trunc	= typeof Math.trunc	== "function"	? Math.trunc			// integral part of x, removing any fractional digits
		: function(x){return HRmath.sign(x)*HRmath.floor(HRmath.abs(x))};
//
HRmath.hypot	= typeof Math.hypot	== "function"	? Math.hypot			// square root of sum of squares of all arguments
		: function(){for(var t=0,n=arguments.length,i=0;i<n;i++){var a=arguments[i];t+=a*a};return HRmath.sqrt(t)};
// =====================================================================================================================
// Not implementing the next three new ECMA V6 Methods:
//	HRmath.clz32	= Math.clz32;	// Returns the number of leading zeroes of a 32-bit integer.
//	HRmath.fround	= Math.fround;	// Returns the nearest single precision float representation of a number.
//	HRmath.imul	= Math.imul;	// Returns the result of a 32-bit integer multiplication.
// #####################################################################################################################
// Other useful Properties:
// ------------------------
HRmath.TAU	= 2*HRmath.PI;				// 2*pi
HRmath.PHI	= 0.5*(HRmath.sqrt(5)+1);		// Golden Ratio
HRmath.D2R	= HRmath.PI/180;			// multiplication factor for degrees to radians conversion
HRmath.R2D	= 180/HRmath.PI;			// multiplication factor for radians to degrees conversion
HRmath.SQD2SR	= (HRmath.PI*HRmath.PI)/(180*180);	// multiplication factor for square degrees to steradians conversion
HRmath.SR2SQD	= (180*180)/(HRmath.PI*HRmath.PI);	// multiplication factor for steradians to square degrees conversion
HRmath.GOOGOL	= 1e100;				// a 1 followed by 100 zeroes
// =====================================================================================================================
// Other useful Methods:
// ---------------------
HRmath.ln	= HRmath.log;							// natural logarithm of x
HRmath.logB	= function(x,b)	{return HRmath.log(x)/HRmath.log(b)};		// base-b logarithm of x
//
HRmath.root	= function(x,y)	{return HRmath.pow(x,arguments.length<2?0.5:1/y)};
HRmath.square	= function(x)	{return x*x};
HRmath.cube	= function(x)	{return x*x*x};
//
HRmath.randomX	= function(x)	{return arguments.length < 1 ? HRmath.random() : HRmath.random()*x};
HRmath.ceilX	= function(x,u)	{return arguments.length < 2 ? HRmath.ceil (x) : HRmath.ceil (x/u)*u};
HRmath.floorX	= function(x,u)	{return arguments.length < 2 ? HRmath.floor(x) : HRmath.floor(x/u)*u};
HRmath.roundX	= function(x,u)	{return arguments.length < 2 ? HRmath.round(x) : HRmath.round(x/u)*u};
HRmath.mod	= function(x,y)	{return (x%y+y)%y};
// =====================================================================================================================
// Circles and spheres:
// --------------------
HRmath.circPerim =
HRmath.circumf	= function(r)	{return HRmath.TAU*r};
HRmath.circArea	= function(r)	{return HRmath.PI*r*r};
HRmath.sphArea	= function(r)	{return 4*HRmath.PI*r*r};
HRmath.sphVol	= function(r)	{return 4*HRmath.PI*r*r*r/3};
// =====================================================================================================================
// Trigonometry:
// -------------
HRmath.sind	= function(x)	{return HRmath.sin(x*HRmath.D2R)};
HRmath.cosd	= function(x)	{return HRmath.cos(x*HRmath.D2R)};
HRmath.tand	= function(x)	{return HRmath.tan(x*HRmath.D2R)};
HRmath.asind	= function(x)	{return HRmath.asin.apply(null,arguments)*HRmath.R2D};
HRmath.acosd	= function(x)	{return HRmath.acos.apply(null,arguments)*HRmath.R2D};
HRmath.atand	= function(x)	{return HRmath.atan.apply(null,arguments)*HRmath.R2D};
HRmath.atan2d	= function(y,x)	{return HRmath.atan2.apply(null,arguments)*HRmath.R2D};
HRmath.hav	= function(x)	{return 0.5*(1-HRmath.cos(x))};			// haversine of angle in radians
HRmath.havd	= function(x)	{return 0.5*(1-HRmath.cosd(x))};		// haversine of angle in degrees
// =====================================================================================================================
// Angle conversions:
// ------------------
HRmath.dms2x	= function(d,m,s){var N=arguments.length,x=HRmath.abs(d)+((N>1?m:0)+(N>2?s:0)/60)/60;return d<0?-x:x};
HRmath.x2dms	= function(x)	{var a=HRmath.abs(x),e=HRmath.floor(a),f=a-e,m=f*60;return[x<0?-e:e,HRmath.floor(m),(m%1)*60]};
HRmath.ms2fr	= function(m,s)	{var N=arguments.length;return ((N>1?m:0)+(N>2?s:0)/60)/60;};
HRmath.fr2ms	= function(x)	{return HRmath.x2dms(x).slice(1)};
// =====================================================================================================================
// Angular distance on a shere in radians (bothe the arguments and the return value).
// Uses equatorial coordinates latitude and longitude.
// If working in spherical coordinates specify pi/2-theta for latitudes latN:
// --------------------------------------------------------------------------
HRmath.greatCircleArc = function(lat1,lng1,lat2,lng2)
{	var deltaLng = HRmath.abs(lng1-lng2), cosDeltaLng = HRmath.cos(deltaLng), sinDeltaLng = HRmath.sin(deltaLng);
	var cosLat1 = HRmath.cos(lat1), cosLat2 = HRmath.cos(lat2), sinLat1 = HRmath.sin(lat1), sinLat2 = HRmath.sin(lat2);
	var p = cosLat2*sinDeltaLng, q = cosLat1*sinLat2 - sinLat1*cosLat2*cosDeltaLng;
	return HRmath.atan2(HRmath.sqrt(p*p+q*q),sinLat1*sinLat2+cosLat1*cosLat2*cosDeltaLng);
};
// ---------------------------------------------------------
// And in degrees (both the arguments and the return value).
// If working in spherical coordinates specify 90-theta for latitudes Bn:
// ----------------------------------------------------------------------
HRmath.greatCircleArcD = function(B1,L1,B2,L2)
{	var d2r = HRmath.D2R;
	return HRmath.greatCircleArc(B1*d2r,L1*d2r,B2*d2r,L2*d2r)*HRmath.R2D;
};
// =====================================================================================================================
// The factorial of n:
// -------------------
HRmath.fac =
HRmath.factorial = function(n)
{	if (isNaN(n) || n < 0)	return NaN;
	if (!isFinite(n))	return Number.POSITIVE_INFINITY;
	if (n%1 != 0)		return NaN;
	if (n > 1e3)		// Stirling approximation for great values of n:
		var f = HRmath.sqrt(HRmath.TAU*n)*HRmath.pow(n/HRmath.E,n);
	else	for (var f = 1, k = n; k > 0; f *= k--);
	return f;
};
// =====================================================================================================================
// Number set functions, any number of arguments can be supplied:
// --------------------------------------------------------------
HRmath.gcd = function()
{	for (var r = NaN, i = 0; i < arguments.length; i++)
	{	var x = HRmath.round(HRmath.abs(arguments[i]));
		if (x == 0) continue;
		if (isNaN(r)) r = x;
		else
		{	var y = HRmath.min(r,x);
			r = HRmath.gcd(y,HRmath.max(r,x)%y);
		};
	};
	return r;
};
HRmath.lcm = function()
{	var gcd = HRmath.gcd.apply(null,arguments);
	for (var p = 1, i = 0; i < arguments.length; i++)
	{	var x = HRmath.round(HRmath.abs(arguments[i]));
		if (x != 0) p *= x;
	};
	return p/gcd;
};
HRmath.rms = function()
{	for (var t = 0, n = arguments.length, i = 0; i < n; i++) {var a = arguments[i]; t += a*a};
	return HRmath.sqrt(t/n);
};
// =====================================================================================================================
// Statistics, any number of arguments can be supplied:
// ----------------------------------------------------
HRmath.stats = function()
{	for (var values = [], argumentsIndex = 0, L = arguments.length; argumentsIndex < L; argumentsIndex++) values = values.concat(arguments[argumentsIndex]);
	for (var i = 0, k = 1, nv = values.length, mu = 0, Q = 0, muh = 0, a = [], cumlog = 0, min = +Infinity, max = -Infinity; i < nv; i++)
	{	var xk = values[i], hk = 1/xk, d = xk-mu;
		if (isFinite(xk))
		{	mu += d/k; Q += d*(xk-mu); muh += (hk-muh)/k;
	 		cumlog += HRmath.log(xk);
	 		a.push(xk);
			min = HRmath.min(xk,min);
			max = HRmath.max(xk,max);
			k++;
		};
	};
	var n = a.length;
	a.sort(function(a,b){return a < b ? -1 : a > b ? 1 : 0});
	var L2 = 0.5*n, L2c = HRmath.ceil(L2);
	return	{mu:		mu
		,sigma:		HRmath.sqrt(Q/n)
		,sigma1:	HRmath.sqrt(Q/(n-1))
		,median:	L2 != L2c ? a[L2c] : 0.5*(a[L2]+a[L2+1])
		,min:		min
		,max:		max
		,geoMean:	HRmath.exp(cumlog/n)
		,harMean:	1/muh
		,count:		n
		,sorted:	a
		,valueOf:	function()
				{	return	{mu	:this.mu
						,sigma	:this.sigma
						,sigma1	:this.sigma1
						,median	:this.median
						,min	:this.min
						,max	:this.max
						,geoMean:this.geoMean
						,harMean:this.harMean
						,count	:this.count
						}
				}
		,toString:	function(){return HRmath.obj2str.apply(this,arguments)}
		};
};
// =====================================================================================================================
// (Cumulative) (Standard) Normal Distribution stuff and inverses thereof:
// -----------------------------------------------------------------------
HRmath.snd	= function(z)		{return	HRmath.exp(-0.5*z*z)/(HRmath.sqrt(2*HRmath.PI))};		// returns probability density
HRmath.csnd	= function(z)		{return	HRmath.min(HRmath.max(0.5*(1+HRmath.erf(z/HRmath.SQRT2)),0),1)};// returns fractile
HRmath.csndinv	= function(f)		{return	HRmath.inverse(HRmath.csnd,f)};					// f = fractile, result = z-score
HRmath.nd	= function(x,mu,sigma)	{return	HRmath.snd (HRmath.x2z(x,mu,sigma))};				// returns probability density
HRmath.cnd	= function(x,mu,sigma)	{return	HRmath.csnd(HRmath.x2z(x,mu,sigma))};				// returns fractile
HRmath.cndinv	= function(f,mu,sigma)	{return HRmath.z2x(HRmath.inverse(HRmath.csnd,f),mu,sigma)};		// f = fractile, result = x
HRmath.x2z	= function(x,mu,sigma)	{return	(x-mu)/sigma};
HRmath.z2x	= function(z,mu,sigma)	{return	mu+z*sigma};
HRmath.z2fractile	= function(z)	{return HRmath.csnd	(z)};
HRmath.fractile2z	= function(f)	{return HRmath.csndinv	(f)};
HRmath.z2fract1sided	= function(z)	{return HRmath.csnd   	(z)-0.5};
HRmath.fract1sided2z	= function(f)	{return HRmath.csndinv	(f+0.5)};
HRmath.z2fract2sided	= function(z)	{return (HRmath.csnd	(z)-0.5)*2};
HRmath.fract2sided2z	= function(f)	{return HRmath.csndinv	(f*0.5+0.5)};
HRmath.iq2z		= function(iq)	{return HRmath.x2z	(iq,100,15)};
HRmath.z2iq		= function(z)	{return HRmath.z2x	(z ,100,15)};
HRmath.iq2fractile	= function(iq)	{return	HRmath.z2fractile(HRmath.iq2z(iq))};
HRmath.fractile2iq	= function(f)	{return	HRmath.z2iq(HRmath.fractile2z(f ))};
// =====================================================================================================================
// The Error Function erf(x) (choice of which erfX to be used is based on extensive testing):
// ------------------------------------------------------------------------------------------
HRmath.erf = function(x)
{	if (x < 0) return -HRmath.erf(-x);
	if (x < 4) return HRmath.erf0(x);
	return HRmath.erf3(x);
};
// =====================================================================================================================
// Different erf approximation algoritms:
// --------------------------------------
// Method returning a value of the error function, calculated using the Maclaurin series
// as given on: https://en.wikipedia.org/wiki/Error_function#Taylor_series
// Date: 2016-03-16
// erf(z) = SUM_n_from_0_to_infinity ( (-1)^n * z^(2n+1) / n! * (2n+1) ) * 2/sqrt(pi):
// -----------------------------------------------------------------------------------
HRmath.erf0 = function(x)
{	if (x < 0) return -HRmath.erf0(-x);
	for (var e=1e-16,m=100000,x2=x*x,n=0,f=1,g=1,y=x,t=x,r=0;n<m&&HRmath.abs(t)>e;r+=t,n++,f*=n,g+=2,y*=-x2,t=y/f/g);
	return n < m ? HRmath.min(HRmath.max(2*r/HRmath.sqrt(HRmath.PI),-1),1) : NaN;
};
// --------------------------------
// and its alternative formulation:
// Date: 2016-03-16
// erf(z) = SUM_n_from_0_to_infinity ( PROD_k_from_1_to_n(-z^2/k) * z/(2n+1) ) * 2/sqrt(pi):
// -----------------------------------------------------------------------------------------
HRmath.erf1 = function(x)
{	if (x < 0) return -HRmath.erf1(-x);
	for (var e = 1e-16, m = 10000, x2 = -x*x, n = 0, n1 = 1, t = 1, r = 0; n < m && HRmath.abs(t) > e; n1 += 2, n++)
	{	for (var k = 1, p = 1; k <= n; p *= x2/k++); t = p*x/n1, r += t;
	};
	return HRmath.min(HRmath.max(2*r/HRmath.sqrt(HRmath.PI),-1),1);
};
// --------------------------------------------------------
// Error Function approximation based on the Brmann series
// as given on: https://en.wikipedia.org/wiki/Error_function#B.C3.BCrmann_series
// Date: 2016-03-16
// (largest relative error (@x=1.3796) = 3.6127e-3):
// -------------------------------------------------
HRmath.erf2 = function(x)
{	if (x < 0) return -HRmath.erf2(-x);
	var x2 = -x*x, ex2 = HRmath.exp(x2), rootpi = HRmath.sqrt(HRmath.PI);
	return HRmath.min(HRmath.max(HRmath.sqrt(1-ex2)*(rootpi/2+31*ex2/200-341*HRmath.exp(2*x2)/8000)*2/rootpi,-1),1);
};
// --------------------------------------------------------------------
// https://en.wikipedia.org/wiki/Error_function#Numerical_approximation
// Date: 2016-03-16
// max error = 1.2e-7:
// -------------------
HRmath.erf3 = function(x)
{	if (x < 0) return -HRmath.erf3(-x);
	var t = 1/(1+x/2);
	return 1-t*HRmath.exp(-x*x-1.26551223+(1.00002368+(0.37409196+(0.09678418+(-0.18628806+(0.27886807
			+(-1.13520398+(1.48851587+(-0.82215223+0.17087277*t)*t)*t)*t)*t)*t)*t)*t)*t);
};
// -------------------------------------------------------------------------------------
// Error Function approximation derived from http://picomath.org/javascript/erf.js.html:
// Date: 2016-03-16
// ----------------
HRmath.erf4 = function(x)
{	if (x < 0) return -HRmath.erf4(-x);
	var a = [0.3275911,0.254829592,-0.284496736,1.421413741,-1.453152027,1.061405429];
	var t = 1/(1+a[0]*x);
	for (var i = 5, y = 0; i > 0; y += a[i], y *= t, i--);
	return 1-y*HRmath.exp(-x*x);
};
// =====================================================================================================================
// Method to get inverse of y=f(x) using bisection, f MUST be continuous and motononically nondecreasing:
// ------------------------------------------------------------------------------------------------------
HRmath.inverse = function(f,y,xmin,xmax)
{	var xlo = (arguments.length > 2 ? xmin : -Number.MAX_VALUE), fxlo = f(xlo);
	if (y <  fxlo) return -Infinity;
	if (y == fxlo) return xlo;
	var xhi = (arguments.length > 3 ? xmax :  Number.MAX_VALUE), fxhi = f(xhi);
	if (y >  fxhi) return Infinity;
	if (y == fxhi) return xhi;
	var M = 1100;	// max. no. of iterations to prevent race condition (base-2 log of Number.MAX_VALUE = 1024)
	for (var d = 0, e = HRmath.pow(2,-51), n = 0; n < M; n++)	// bisection with max. no of iterations
	{	var x0 = x, x = 0.5*xlo+0.5*xhi;			// average with highest accuracy, so not: (xlo+xhi)/2
		if (x == x0 || HRmath.abs(xhi-xlo) < e) return x;
		var fx = f(x);
		if (fx <= y) xlo = x;
		if (fx >= y) xhi = x;
	};
	return NaN;
};
// =====================================================================================================================
// Integration.
// --------------------------------------------
// Simpson's rule for a number of subintervals:
// --------------------------------------------
HRmath.integrateN=function(x0,x1,fx,n){for(var r=0,d=x1-x0,a=x0,b=a+d/n,i=1;i<=n;a=b,b=x0+d*(++i)/n)r+=(fx(a)+4*fx(0.5*a+0.5*b)+fx(b))*(b-a)/6;return r};
// -----------------------------------------------------------------------------------------
// Calculating subinterval count from a size (count will be rounded, so dx is approximated):
// -----------------------------------------------------------------------------------------
HRmath.integrateDx=function(x0,x1,fx,dx){return HRmath.integrateN(x0,x1,fx,HRmath.round((x1-x0)/dx))};
// -----------------------------------------------------------------
// Default integration method with an arbitrary no. of subintervals:
// -----------------------------------------------------------------
HRmath.integrate=function(x0,x1,fx){return HRmath.integrateN(x0,x1,fx,1000*HRmath.sqrt(x1-x0))};
// =====================================================================================================================
// Axb solves A.x = b, where A is an NxN matrix and b an N-vector
// It takes the largest parts from A and b for which a valid N is possible
// It "wipes" the matrix using complete pivoting
// Return value:
//	{x:	vector x as an array
//	,q:	quality factor = determinant of A divided by nth power of mean value of all elements in A and b
//	,A:	copy of input matrix actually used
//	,b:	copy of input vector actually used
//	}
// ------
HRmath.Axb = function(A,b)
{	// determine square matrix:
	var nr = HRmath.min(A.length,b.length);
	for (var nc = A[0].length, ir = 1; ir < nr; ir++) nc = HRmath.min(nc,A[ir].length);
	var N = HRmath.min(nr,nc);
	// copy relevant part of A and b, and set up list of column numbers (because of pivoting, have to find back correct index no. of x):
	for (var A0 = [], AA = [], b0 = [], bb = [], cn = [], ir = 0; ir < N; ir++)
	{	cn[ir] = ir;	// ir is NOT an error since cn is one-dimensional and matrix is square
		b0[ir] = +b[ir];
		bb[ir] = +b[ir];
		A0[ir] = [];
		AA[ir] = [];
		for (var ic = 0; ic < N; ic++)
		{	A0[ir][ic] = +A[ir][ic];
			AA[ir][ic] = +A[ir][ic];
	};	};
	// mean value of NxN matrix elements for later determinant evaluation:
	for (var mu = 0, k = 1, ir = 0; ir < N; ir++) for (var ic = 0; ic < N; ic++, k++) mu += (AA[ir][ic]-mu)/k;
	for (var ir = 0; ir < N; ir++, k++) mu += (bb[ir]-mu)/k;
	// wipe the matrix to make lower left all nought (LU decomposition with complete pivoting):
	for (var det = 1, id = 0; id < N; id++)
	{	// find pivot = maximum value in lower right rest of AA:
		var pRow = id, pCol = id, pVal = AA[pRow][pCol];
		for (var ir = id+1; ir < N; ir++)
		{	for (var ic = id+1; ic < N; ic++)
			{	if (AA[ir][ic] > pVal) {pRow = ir, pCol = ic, pVal = AA[pRow][pCol]}
		};	};
		// exchange rows:
		for (var ic = id; ic < N; ic++) {var tmp = AA[id][ic]; AA[id][ic] = AA[pRow][ic]; AA[pRow][ic] = tmp};
		var tmp = bb[id]; bb[id] = bb[pRow]; bb[pRow] = tmp;
		// exchange columns:
		for (var ir = id; ir < N; ir++) {var tmp = AA[ir][id]; AA[ir][id] = AA[ir][pCol]; AA[ir][pCol] = tmp};
		var tmp = cn[id]; cn[id] = cn[pCol]; cn[pCol] = tmp;
		// wipe = subtract [current row] times [correct factor] from all lower rows, zeroing lower part of current column
		for (var ir = id+1; ir < N; ir++)
		{	var factor = AA[ir][id]/pVal; AA[ir][id] = 0;
			for (var ic = id+1; ic < N; ic++) AA[ir][ic] -= factor*AA[id][ic];
			bb[ir] -= factor*bb[id];
		};
		det *= pVal;
	};
	// find x such that AA.x = bb:
	for (var x = [], id = N; (id--) > 0;)
	{	for (var tmp = 0, ic = N; (--ic) > id;) tmp += AA[id][ic]*x[cn[ic]];
		x[cn[id]] = (bb[id]-tmp)/AA[id][id];
	};
	var q = det/HRmath.pow(mu,N);		// quality/reliability indicator
	var result = {x:x,q:q,A:A0,b:b0};	// and that's it
	result.valueOf	=
	result.toString	= function(){return HRmath.obj2str.apply(this,arguments)};
	return result;
};
// =====================================================================================================================
// getDataPoints parses its arguments as follows:
//	0: argument count = 1: an array of length = 2: elements  are: [x0,x1,x2,x3...] and [y0,y1,y2,y3...]
//	1: argument count = 1: an array of length > 2: elements  are: all [x,y] pairs (as two-element arrays)
//	2: argument count = 2:                         arguments are: [x0,x1,x2,x3...] and [y0,y1,y2,y3...]
//	3: argument count > 2:                         arguments are: all [x,y] pairs
// It returns an object with two arrays:
//	{xy: all [x,y] pairs sorted by ascending x,y
//	,yx: the transposed data point set = all [y,x] pairs sorted by ascending y,x
//	}
// Only valid data points are returned i.e. both x and y must be given and be interpretable as a finite Number
// Usage in another function:
//	var dataPoints = HRmath.getDataPoints.apply(null,arguments);
// -----------------------------------------------------------------------------------------------------------
HRmath.getDataPoints = function()
{	var e, xy = [], yx = [], N = NaN;
	switch (arguments.length)
	{	case 0:	return null;
		case 1:	if ("length" in arguments[0]) return HRmath.getDataPoints.apply(null,arguments[0]);
			break;
		case 2:	var N = HRmath.min
				("length" in arguments[0] ? +arguments[0].length : NaN
				,"length" in arguments[1] ? +arguments[1].length : NaN
				);
			var modus = 0;							// arguments are: [x0, x1, ...], [y0, y1, ...]
			break;
		default:
			var N = arguments.length;
			var modus = 1;							// arguments are: [x0, y0], [x1, y1], ...
	};
	if (isFinite(N)) for (var i = 0; i < N; i++)
	{	var xi = NaN, yi = NaN;
		try {switch (modus)
			{	case 0: xi = +arguments[0][i]; yi = +arguments[1][i]; break;	// plus signs convert the value to Number if necs
				case 1: xi = +arguments[i][0]; yi = +arguments[i][1]; break;
			};
			if (isFinite(xi) && isFinite(yi)) {xy.push([xi,yi]); yx.push([yi,xi])};	// data points and transposed data points
		}catch(e){}finally{};
	};
	xy.sort(function(a,b){return a[0]<b[0]?-1:a[0]>b[0]?1:a[1]<b[1]?-1:a[1]>b[1]?1:0});	// sort [x,y] pairs by ascending x, then y
	yx.sort(function(a,b){return a[0]<b[0]?-1:a[0]>b[0]?1:a[1]<b[1]?-1:a[1]>b[1]?1:0});	// sort transposed data points similarly
	var result = {xy:xy,yx:yx};
	result.valueOf	=
	result.toString	= function(){return HRmath.obj2str.apply(this,arguments)};
	return result;
};
// =====================================================================================================================
// function getStraightLine uses getDataPoints to parse its arguments
// and it returns an object with:
//	muPhi:		the average angle of all straight lines through:
//				if HRmath.getStraightLineDoAll then all possible pairs of data points
//				else all consecutive pairs of data points
//	sigmaPhi:	the standard deviation thereof, this will be zero if the data points are on a perfectly straight line
//	steep:		true/false to indicate if the average angle is above 45 degrees
//	muX:		the average x-value of all valid data points
//	muY:		the average y-value of all valid data points
//	dataSet:	array of all valid finite data points actually used as [x,y] pairs
//	didAll:		true/false value of HRmath.getStraightLineDoAll
//	toString:	guess what?
//	valueOf:	identical to toString
//		[muX,muY,muPhi] fully determine a best-fit straight line through the data points
//		sigmaPhi indicates the "unstraightness", the closer to nought, the better the closer the data points are to a straight line
// ---------------------------------------------------
HRmath.getStraightLineDoAll = false;
HRmath.getStraightLine = function()
{	var dataPoints = HRmath.getDataPoints.apply(null,arguments), xy = dataPoints.xy, yx = dataPoints.yx;
	var mupn = 0, mupna = 0, mupt = 0, mupta = 0, Qn = 0, Qt = 0, mux = 0, muy = 0;
	for (var i = 0, k = 1, n = 1; i < xy.length; i++, k++)
	{	var xin = xy[i][0], yin = xy[i][1]; mux += (xin-mux)/k, muy += (yin-muy)/k;
		var yit = yx[i][0], xit = yx[i][1];
		for (var j = i+1; j < xy.length; j++, n++)
		{	var xjn = xy[j][0], yjn = xy[j][1];
			var yjt = yx[j][0], xjt = yx[j][1];
			var dxn = xjn-xin, dyn = yjn-yin, phin = HRmath.atan2(dyn,dxn), d = phin-mupn;
			var dyt = yjt-yit, dxt = xjt-xit, phit = HRmath.atan2(dxt,dyt), d = phit-mupt;
			mupn += d/n, Qn += d*(phin-mupn), mupna += (HRmath.abs(phin)-mupna)/n;
			mupt += d/n, Qt += d*(phit-mupt), mupta += (HRmath.abs(phit)-mupta)/n;
			if (!HRmath.getStraightLineDoAll) j = xy.length;	// force end of j-loop but still do the n++
	};	};
	var result =
		(mupna > mupta
		? {muPhi: HRmath.PI/2-mupt, sigmaPhi: HRmath.sqrt(Qt/--n), steep: true}
		: {muPhi:             mupn, sigmaPhi: HRmath.sqrt(Qn/--n), steep: false}
		);
	result.muX	= mux;
	result.muY	= muy;
	result.dataSet	= xy;
	result.didAll	= HRmath.getStraightLineDoAll ? true : false;
	result.valueOf	=
	result.toString	= function(){return HRmath.obj2str.apply(this,arguments)};
	return result;
};
// =====================================================================================================================
// fitPolynomial determines the coefficients of a polynomial through the data points
// Argument interpretation is identical to how getStrainghtLine does it, but the transposed set is ignored
// The 'steep' and 'dataSet' properties of 'getDataPoints' are copied to the result
// --------------------------------------------------------------------------------
HRmath.fitPolynomial = function()
{	var dataPoints = HRmath.getDataPoints.apply(null,arguments), xy = dataPoints.xy;
	for (var A = [], b = [], i = 0; i < xy.length; i++)
	{	A[i] = []; b[i] = xy[i][1];
		for (var xi = xy[i][0], Aij = 1, j = 0; j < xy.length; j++) {A[i].push(Aij); Aij *= xi};
	};
	var result = HRmath.Axb(A,b);
	result.steep = dataPoints.steep;
	result.dataSet = dataPoints.dataSet;
	result.valueOf	=
	result.toString	= function(){return HRmath.obj2str.apply(this,arguments)};
	return result;
};
// =====================================================================================================================
// Function for use as toString of an Object
// Usage: obj.toString = function(){return HRmath.obj2str.apply(this,arguments)};
// ------------------------------------------------------------------------------
HRmath.obj2str = function()
{	var x = [];
	var savedToString = Array.prototype.toString;
	Array.prototype.toString = function(){return "["+this.join(",")+"]"};
	for (var p in this) if(this.hasOwnProperty(p) && typeof this[p] != "function") x.push(p.concat(":",this[p]));
	Array.prototype.toString = savedToString;
	return "{"+x.join(",")+"}";
};
// #####################################################################################################################
// [EOF] - Copyright (c) 2017 Henk Reints, http://henk-reints.nl
