// #################################################################################################
// File:	readCsv.js
// Copyright:	(c) 2018, Henk Reints, http://henk-reints.nl
// Description:	package to read a CSV file without any special characters in the fields
//		the very first line of the file is assumed to be a list of field names
// #################################################################################################
// next is a general purpose file reading function
// it reads an entire input file and for each line read, callbackFunction is invoked with inputLine,lineNo and all given extra arguments, which can be any number of arguments
// if they are objects then JavaScripts calls them by reference, so you can save data in an Array or so
// the 'fso' argument should provide the File System Object
// --------------------------------------------------------
function readFile(fso,fileName,callbackFunction)
{	var extraArgumentsArray = Array.prototype.slice.call(arguments,readFile.length);
	var inStream = fso.OpenTextFile(fileName,1);	//1 = ForReading
	for (var lineNo = 0; !inStream.AtEndOfStream; lineNo++) callbackFunction.apply(null,[inStream.ReadLine(),lineNo].concat(extraArgumentsArray));
	inStream.Close();
};
// ================================================================================================
// next function parses each line read and stores it in 'dataList', which should be an Array object
// it should not contain any special characters, splitting into fields is done at every comma without interpretation
// the very first line of the file is assumed to be a list of field names
// 'filterFunction' is an optional argument that is invoked for each record just before it is being pushed to 'dataList',
// the latter then only occurs if 'filterFunction' returns something that is evaluated as TRUE by the if-statement
// if used as a callback function for 'readFile' as defined above,
// then the result is dataList being a list of records,
// where each record is an object with property names equal to the field names in the very first line
// --------------------------------------------------------------------------------------------------
var autoFieldNamePrefix		= "field";
var indexNamePrefix		= "index";
var inputRecordNumberFieldName	= "inputRecordNumber";
function handleCSVinputLine(inputLine,lineNo,dataList,filterFunction)
{	var dontHaveFilterFunction = (typeof filterFunction != "function");
	var fieldValueList = inputLine.split(",");
	var fieldValueListLength = fieldValueList.length;
	if (lineNo > 0)
	{	var fieldNameList = dataList.fieldNameList;
		var record = {};
		record[fieldNameList[0]] = lineNo;	// the 'inputRecordNumber' field
		for (var fieldValueListIndex = 0, fieldNameListIndex = 1; fieldValueListIndex < fieldValueListLength; fieldValueListIndex++, fieldNameListIndex++)
		{	var fieldName  = fieldNameList[fieldNameListIndex];
			var fieldValue = fieldValueList[fieldValueListIndex];
			while (fieldNameList.length <= fieldNameListIndex) fieldNameList.push(autoFieldNamePrefix+fieldNameList.length);
			record[fieldName] = fieldValue.replace(/^\s+|\s+$/g,"");
		};
		fieldNameListLength = fieldNameList.length;
		for (; fieldNameListIndex < fieldNameListLength; fieldNameListIndex++)
		{	var fieldName =  fieldNameList[fieldNameListIndex];
			record[fieldName] = "";
		};
		record.toString = dataList.recordToString;
		if (dontHaveFilterFunction || filterFunction(record)) dataList.push(record);
	} else
	{	dataList.length = 0;	// this clears the Array, thus removing any old stuff
		dataList.fieldNameList = [inputRecordNumberFieldName];
		var fieldNameList = dataList.fieldNameList;
		for (var fieldValueListIndex = 0; fieldValueListIndex < fieldValueListLength; fieldValueListIndex++)
		{	var fieldValue = fieldValueList[fieldValueListIndex].replace(/^\s+|\s+$/g,"");	// leading or trailing white space is trimmed
			fieldNameList.push(fieldValue != "" ? fieldValue : autoFieldNamePrefix+fieldNameList.length-1);
		};
		dataList.recordToString = function()
		{	var result = [];
			for (var propertyName in this) if (this.hasOwnProperty(propertyName)) result.push([propertyName,this[propertyName]].join(":"));
			return "{"+result+"}";
		};
		dataList.toString = function()
		{	var result = [];
			for (var index = 0; index < this.length; index++) result.push(this[index].toString());
			return "["+result.join("\n,")+"\n]";
		};
	};
};
// ==============================================================================================================================================
// next function assumes the very first line is a white-space separated ruler with field names, all other lines are split according to this ruler
// ----------------------------------------------------------------------------------------------------------------------------------------------
function handleFixedFieldWidthInputLine(inputLine,lineNo,dataList,filterFunction)
{	var dontHaveFilterFunction = (typeof filterFunction != "function");
	if (lineNo > 0)
	{	var fieldNameList	= dataList.fieldNameList;
		var fieldPositionList	= dataList.fieldPositionList;
		var fieldNameListLength = fieldNameList.length;
		var fieldPositionListMaxIndex = fieldPositionList.length - 1;
		var record = {};
		record[fieldNameList[0]] = lineNo;	// the 'inputRecordNumber' field
		for (fieldNameListIndex = 1; fieldNameListIndex < fieldNameListLength; fieldNameListIndex++)
		{	var fieldName     = fieldNameList[fieldNameListIndex];
			var fieldPosition = fieldPositionList[fieldNameListIndex];
			var fieldWidth    = fieldNameListIndex < fieldPositionListMaxIndex ? fieldPositionList[fieldNameListIndex+1]-fieldPosition : NaN;
			var fieldValue    = isFinite(fieldWidth) ? inputLine.substr(fieldPosition,fieldWidth) : inputLine.slice(fieldPosition);
			record[fieldName] = fieldValue.replace(/^\s+|\s+$/g,"");
		};
		record.toString = dataList.recordToString;
		if (dontHaveFilterFunction || filterFunction(record)) dataList.push(record);
	} else
	{	dataList.length = 0;	// this clears the Array, thus removing any old stuff
		dataList.fieldNameList = [inputRecordNumberFieldName];
		dataList.fieldPositionList = [NaN];
		var nonWhiteRegExp = /\S+/g;
		while (true)
		{	var match = nonWhiteRegExp.exec(inputLine);
			if (match == null) break;
			dataList.fieldNameList.push(match[0]);
			dataList.fieldPositionList.push(match.index);
		};
		dataList.recordToString = function()
		{	var result = [];
			for (var propertyName in this) if (this.hasOwnProperty(propertyName)) result.push([propertyName,this[propertyName]].join(":"));
			return "{"+result+"}";
		};
		dataList.toString = function()
		{	var result = [];
			for (var index = 0; index < this.length; index++) result.push(this[index].toString());
			return "["+result.join("\n,")+"\n]";
		};
	};
};
// ================================================
// next function transposes the result of the above
// ------------------------------------------------
function transposeDataList(dataList)
{	var transposed = {};
	var fieldNameList = dataList.fieldNameList;
	for (var fieldNameListIndex = 0, fieldNameListLength = fieldNameList.length; fieldNameListIndex < fieldNameListLength; fieldNameListIndex++)
	{	var fieldName = fieldNameList[fieldNameListIndex];
		transposed[fieldName] = [];
		for (var dataListIndex = 0, dataListLength = dataList.length; dataListIndex < dataListLength; dataListIndex++) transposed[fieldName].push(dataList[dataListIndex][fieldName]);
	};
	transposed.toString = function()
	{	var result = [];
		for (var propertyName in this) if (this.hasOwnProperty(propertyName)) result.push([propertyName,"["+this[propertyName]+"]"].join(":"));
		return "{"+result.join("\n,")+"\n}";
	};
	dataList.transposed = transposed;
	return transposed;
};
// =========================================================================================================================================================================
// next function sorts 'dataList' by any number of field names in the order given (can also be specified in one or more Arrays, default is the already present fieldNameList
// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------
function sortDataList(dataList)
{	var sortByFieldNamesList = arguments.length > sortDataList.length ? Array.prototype.slice.call(arguments,sortDataList.length).toString().split(",") : dataList.fieldNameList;

	var sortByFieldNamesListLength = sortByFieldNamesList.length;
	dataList.sort(function(a,b)
	{	for (var sortByFieldNamesListIndex = 0; sortByFieldNamesListIndex < sortByFieldNamesListLength; sortByFieldNamesListIndex++)
		{	var sortByFieldName = sortByFieldNamesList[sortByFieldNamesListIndex];
			if (a[sortByFieldName] < b[sortByFieldName]) return -1;
			if (a[sortByFieldName] > b[sortByFieldName]) return  1;
		};
		return 0;
	});
	return dataList;
};
// ================================================================================================================================================
// next function adds an index to 'dataList' using one of the fields as key, in case of multiple occurrance the last record encountered will remain
// the result is more an associative array than a very efficient index, it just builds an object with the key values as property names
// it will be added to dataList as a property named indexNamePrefix+keyFieldName
// -----------------------------------------------------------------------------
function buildIndex(dataList,keyFieldName)
{	var result = {};
	for (var dataListIndex = 0, dataListLength = dataList.length; dataListIndex < dataListLength; dataListIndex++)
	{	var record = dataList[dataListIndex];
		var keyValue = record[keyFieldName];
		result[keyValue] = dataListIndex;
	};
	result.dataList = dataList;
	result.toString = function()
	{	var result = [];
		var dataList = this.dataList;
		for (var propertyName in this) if (this.hasOwnProperty(propertyName) && propertyName != "dataList" && propertyName != "toString")
		{	var dataListIndex = this[propertyName];
			var record = dataList[dataListIndex];
			result.push(propertyName+":{index:"+dataListIndex+",record:"+record+"}");
		};
		return "{"+result.join("\n,")+"\n}";
	};
	dataList[indexNamePrefix+keyFieldName] = result;
	return result;
};
// #################################################################################################
// end of file - readCsv.js - Copyright (c) 2018, Henk Reints, http://henk-reints.nl
